@@ -1,10 +1,9 @@
 from __future__ import unicode_literals
-import datetime
 import fcntl
 import os
-import random
 import signal
 import threading
+import time
 
 from prompt_toolkit.terminal.vt100_input import InputStream
 from prompt_toolkit.utils import DummyContext, in_main_thread
@@ -14,21 +13,22 @@ from .callbacks import EventLoopCallbacks
 from .inputhook import InputHookContext
 from .posix_utils import PosixStdinReader
 from .utils import TimeIt
-from .select import select_fds
+from .select import AutoSelector, Selector, fd_to_int
 
 __all__ = (
     'PosixEventLoop',
 )
 
-_now = datetime.datetime.now
+_now = time.time
 
 
 class PosixEventLoop(EventLoop):
     """
     Event loop for posix systems (Linux, Mac os X).
     """
-    def __init__(self, inputhook=None):
+    def __init__(self, inputhook=None, selector=AutoSelector):
         assert inputhook is None or callable(inputhook)
+        assert issubclass(selector, Selector)
 
         self.running = False
         self.closed = False
@@ -37,6 +37,7 @@ class PosixEventLoop(EventLoop):
 
         self._calls_from_executor = []
         self._read_fds = {} # Maps fd to handler.
+        self.selector = selector()
 
         # Create a pipe for inter thread communication.
         self._schedule_pipe = os.pipe()
@@ -93,18 +94,21 @@ class PosixEventLoop(EventLoop):
         with ctx:
             while self._running:
                 # Call inputhook.
-                with TimeIt() as inputhook_timer:
-                    if self._inputhook_context:
+                if self._inputhook_context:
+                    with TimeIt() as inputhook_timer:
                         def ready(wait):
                             " True when there is input ready. The inputhook should return control. "
                             return self._ready_for_reading(current_timeout[0] if wait else 0) != []
                         self._inputhook_context.call_inputhook(ready)
+                    inputhook_duration = inputhook_timer.duration
+                else:
+                    inputhook_duration = 0
 
                 # Calculate remaining timeout. (The inputhook consumed some of the time.)
                 if current_timeout[0] is None:
                     remaining_timeout = None
                 else:
-                    remaining_timeout = max(0, current_timeout[0] - inputhook_timer.duration)
+                    remaining_timeout = max(0, current_timeout[0] - inputhook_duration)
 
                 # Wait until input is ready.
                 fds = self._ready_for_reading(remaining_timeout)
@@ -122,17 +126,23 @@ class PosixEventLoop(EventLoop):
                     # case.
                     tasks = []
                     low_priority_tasks = []
-                    now = _now()
+                    now = None  # Lazy load time. (Fewer system calls.)
 
                     for fd in fds:
                         # For the 'call_from_executor' fd, put each pending
                         # item on either the high or low priority queue.
                         if fd == self._schedule_pipe[0]:
                             for c, max_postpone_until in self._calls_from_executor:
-                                if max_postpone_until is None or max_postpone_until < now:
+                                if max_postpone_until is None:
+                                    # Execute now.
                                     tasks.append(c)
                                 else:
-                                    low_priority_tasks.append((c, max_postpone_until))
+                                    # Execute soon, if `max_postpone_until` is in the future.
+                                    now = now or _now()
+                                    if max_postpone_until < now:
+                                        tasks.append(c)
+                                    else:
+                                        low_priority_tasks.append((c, max_postpone_until))
                             self._calls_from_executor = []
 
                             # Flush all the pipe content.
@@ -142,10 +152,6 @@ class PosixEventLoop(EventLoop):
                             if handler:
                                 tasks.append(handler)
 
-                    # Handle everything in random order. (To avoid starvation.)
-                    random.shuffle(tasks)
-                    random.shuffle(low_priority_tasks)
-
                     # When there are high priority tasks, run all these.
                     # Schedule low priority tasks for the next iteration.
                     if tasks:
@@ -179,8 +185,7 @@ class PosixEventLoop(EventLoop):
         """
         Return the file descriptors that are ready for reading.
         """
-        read_fds = list(self._read_fds.keys())
-        fds = select_fds(read_fds, timeout)
+        fds = self.selector.select(timeout)
         return fds
 
     def received_winch(self):
@@ -225,15 +230,23 @@ class PosixEventLoop(EventLoop):
         Call this function in the main event loop.
         Similar to Twisted's ``callFromThread``.
 
-        :param _max_postpone_until: `None` or `datetime` instance. For interal
+        :param _max_postpone_until: `None` or `time.time` value. For interal
             use. If the eventloop is saturated, consider this task to be low
             priority and postpone maximum until this timestamp. (For instance,
             repaint is done using low priority.)
         """
+        assert _max_postpone_until is None or isinstance(_max_postpone_until, float)
         self._calls_from_executor.append((callback, _max_postpone_until))
 
         if self._schedule_pipe:
-            os.write(self._schedule_pipe[1], b'x')
+            try:
+                os.write(self._schedule_pipe[1], b'x')
+            except (AttributeError, IndexError, OSError):
+                # Handle race condition. We're in a different thread.
+                # - `_schedule_pipe` could have become None in the meantime.
+                # - We catch `OSError` (actually BrokenPipeError), because the
+                #   main thread could have closed the pipe already.
+                pass
 
     def stop(self):
         """
@@ -257,13 +270,19 @@ class PosixEventLoop(EventLoop):
 
     def add_reader(self, fd, callback):
         " Add read file descriptor to the event loop. "
+        fd = fd_to_int(fd)
         self._read_fds[fd] = callback
+        self.selector.register(fd)
 
     def remove_reader(self, fd):
         " Remove read file descriptor from the event loop. "
+        fd = fd_to_int(fd)
+
         if fd in self._read_fds:
             del self._read_fds[fd]
 
+        self.selector.unregister(fd)
+
 
 class call_on_sigwinch(object):
     """
